<?php

namespace App\Service;

use App\Exception;

class OrderService {
    
    private $util;
    private $productService;
    private $branchFactory;
    private $factory;
    private $paymentFactory;

    public function __construct (\App\Util\Util $util, 
                                 \App\Service\ProductService $productService, 
                                 \App\Factory\BranchFactory $branchFactory, 
                                 \App\Factory\OrderFactory $factory, 
                                 \App\Factory\PaymentFactory $paymentFactory)
    {
        $this->util            = $util;
        $this->productService  = $productService;
        $this->branchFactory   = $branchFactory;
        $this->factory         = $factory;
        $this->paymentFactory  = $paymentFactory;
    }
    
    private function validateDefaultParams($params)
    {

        if(empty($params)) 
            throw new Exception\ParamsException("Nenhum caixa foi encontrado.");
        
        if(empty($params['CDFILIAL'])) 
            throw new Exception\ParamsException("Nenhuma filial encontrada.");
        
        if(empty($params['CDVENDEDOR'])) 
            throw new Exception\ParamsException("Vendedor padrao nao encontrado.");
        
        if(empty($params['CDOPERADOR'])) 
            throw new Exception\ParamsException("O vendedor {$params['NMFANVEN']} nao esta associado a um operador valido.");
        
        if(empty($params['CDLOJA'])) 
            throw new Exception\ParamsException("Loja nao encontrada para o caixa {$params['CDCAIXA']}.");
        
        if(empty($params['NRMESA'])) 
            throw new Exception\ParamsException("Mesa padrão nao encontrada para a loja {$params['CDLOJA']}.");
        
        if(empty($params['CDCLIENTE'])) 
            throw new Exception\ParamsException("Cliente padrão nao encontrado para filial {$params['CDFILIAL']}.");
        
        if(empty($params['CDCONSUMIDOR'])) 
            throw new Exception\ParamsException("Consumidor padrão nao encontrado.");
    }
    
    public function getDefaultParams($cdfilial)
    {
        $defaultParams = $this->branchFactory->getDefaultParams($cdfilial);
        $this->validateDefaultParams($defaultParams);
        return $defaultParams;
    }
    
    private function TypesShippingDiscounts ($type)
    {
        $validTypes = [
            "prime_discount_shipping" , "free_shipping" , "charges_by_value" ,
            "charge_by_value", "charge.shipping" , "charge_by_value.shipping"
        ];
        
        return in_array($type, $validTypes);
    }

    public function openOrderBill ($generalParams, $order, $consumer)
    {
        $nrvendarest = $this->util->newCode("VENDAREST".$generalParams["CDFILIAL"], 10);
        $nrcomanda   = $this->util->newCode("COMANDAVEN".$generalParams["CDFILIAL"], 10);
        $isAutomaticPrint = $this->branchFactory->isPosEnabledToAutomaticOrderPrinting($generalParams["CDFILIAL"], $generalParams["CDLOJA"], $generalParams["CDCAIXA"]);
        
        if(!empty($isAutomaticPrint)){
             $order["IDIMPAUTCUPEXP"] = $isAutomaticPrint["IDIMPAUTCUPEXP"];
             $order["IDIMPAUTCUPFOS"] = $isAutomaticPrint["IDIMPAUTCUPFOS"];
        }
        
        if (($order["IDIMPAUTCUPEXP"] == "S") && ($order["IDIMPAUTCUPFOS"] == "S"))  {
            $order["IDIMPCUPEXP"] = 2;
        } else { 
            $order["IDIMPCUPEXP"] = 0;
        }
        
        $order["NRVENDAREST"] = $nrvendarest;
        $order["NRCOMANDA"]   = $nrcomanda;
        $order["IDSTCOMANDA"] = $generalParams["IDCONTROPROD"] === "S" ? "1" : "3";
        
        $order['DELIVERYMETHOD'] = $this->getDeliveryMethod($order['delivery_method']);

        $isDuplicate = $this->factory->validateDuplication($generalParams, $order["order_id"]);

        // Observação de pagamento online
        $generalParams['DSOBSCOMANDA'] = "";
        
        // Taxa de entrega
        $deliveryFee = 0;
        
        // Desconto na taxa de entrega pela Rappi
        $rappiShippingDiscount = 0;
            
        // Desconto na taxa de serviço pela Rappi
        $rappiServiceFeeDiscount = 0;
        
        // Descontos na comanda pela Rappi
        $rappiOrderDiscount = 0;
        
        if(empty($order["totals"]["total_order"])) 
            throw new Exception\IntegrationException("O pedido " . $order["order_id"] . " não possui valor total.");
            
        if ($isDuplicate)
            throw new Exception\IntegrationException("O pedido " . $order["order_id"] . " já foi integrado pelo sistema.", 302);
       
        if((!empty($order["totals"]["other_totals"]["total_rappi_credits"]) && $order["totals"]["other_totals"]["total_rappi_credits"] > 0) ||
           (!empty($order["totals"]["other_totals"]["total_rappi_pay"]) && $order["totals"]["other_totals"]["total_rappi_pay"] > 0) ||
           ($order["payment_method"] === "cc")){
           
            $generalParams['DSOBSCOMANDA'] .= "PAGO ONLINE";
        }
        
        if($order["totals"]["total_to_pay"] > 0){
            $generalParams['DSOBSCOMANDA'] .= " (RECEBER " . $order['totals']['total_to_pay'] . " EM DINHEIRO)";
        }
        
        if($order["delivery_method"] != "delivery"){
            $order["totals"]["total_order"] -= $order["totals"]["other_totals"]["tip"];
        }


        if($order["delivery_method"] == "delivery"){
            $order["totals"]["total_order"] += $deliveryFee;
        }

        if($order["delivery_method"] != "delivery"){
            $order["totals"]["total_order"] -= $order["totals"]["other_totals"]["tip"];
        }
        
        $generalParams["NRSEQPRODCUP"] = $this->util->newCode("ITCOMANDAVEN".$generalParams["CDFILIAL"], 10);
        $sharedDiscount = 0;

        if(!empty($order["discounts"])) {
            foreach($order["discounts"] as $discount){
                 if($this->TypesShippingDiscounts($discount["type"])) { // DESCONTOS NA TAXA DE ENTREGA
                    if(($discount["amount_by_rappi"] + $discount["amount_by_partner"]) != $discount["value"]){
                        throw new Exception\ParamsException("Há divergência entre o valor total do desconto e o desconto aplicado.");
                    }
                    if($discount["amount_by_rappi"] > 0){
                        $rappiShippingDiscount += $discount["amount_by_rappi"];
                    }
                    if($discount["amount_by_partner"] > 0) {
                        $order["totals"]["charges"]["shipping"] -= $discount["amount_by_partner"];
                    }
                } else if($discount["type"] == "service_fee") { //DESCONTO NA TAXA DE SERVIÇO
                    if($order["delivery_method"] != "delivery") { // O valor do service_fee só é incluido para modalidades marketplace/pickup
                        if(($discount["amount_by_rappi"] + $discount["amount_by_partner"]) != $discount["value"]) {
                            throw new Exception\ParamsException("Há divergência entre o valor total do desconto e o desconto aplicado.");
                        }
                        if($discount["amount_by_rappi"] > 0) {
                            $rappiServiceFeeDiscount += $discount["amount_by_rappi"];
                        }
                        if($discount["amount_by_partner"] > 0) {
                            $order["totals"]["charges"]["service_fee"] -= $discount["amount_by_partner"];
                        }
                    }
                } else {
                    if (!empty($discount["product_id"]) || ($discount["type"] == "offer_by_product" && !empty($discount["product_id"]))) {    // DESCONTO NO VALOR DE UM ITEM OU SUBITEM
                        foreach($order["items"] as &$item) {
                            
                            if(($discount["amount_by_rappi"] + $discount["amount_by_partner"]) != $discount["value"]) {
                                throw new Exception\ParamsException("Há divergência entre o valor total do desconto e o desconto aplicado.");
                            }
                            if($item["id"] == $discount["product_id"]) {
                                
                                if($discount["amount_by_rappi"] > 0) {
                                    $rappiOrderDiscount += $discount["amount_by_rappi"];
                                }
                                if($discount["amount_by_partner"] > 0){
                                    if(!isset($item["discount"]))
                                        $item["discount"] = 0;
                                    $item["discount"] += $discount["discount_product_unit_value"];
                                }
                            }
                            if(!empty($item["subitems"])) {
                                foreach($item["subitems"] as &$subitems) {
                                    if(($discount["amount_by_rappi"] + $discount["amount_by_partner"]) != $discount["value"]) {
                                        throw new Exception\ParamsException("Há divergência entre o valor total do desconto e o desconto aplicado.");
                                    }
                                    
                                    if($subitems["id"] == $discount["product_id"]) {
                                        if($discount["amount_by_rappi"] > 0) {
                                            $rappiOrderDiscount += $discount["amount_by_rappi"];
                                        }
                                        if($discount["amount_by_partner"] > 0) {
                                            $subitems["discount"] = $discount["discount_product_unit_value"];
                                        }
                                    }
                                }
                            }
                        }
                    } else {
                        if($discount["amount_by_rappi"] > 0) {
                                $rappiOrderDiscount += $discount["amount_by_rappi"];
                            }
                        if($discount["amount_by_partner"] > 0) {
                            $sharedDiscount += $discount["value"];
                        }
                    }
                }
            }
        }
        
        //var_dump($order);
        
        if(!empty($order["totals"]["charges"]) && !empty($order["totals"]["charges"]["shipping"])){
            if($order["delivery_method"] != "delivery") {
                $deliveryFee += $order["totals"]["charges"]["shipping"];
            }
        }
            
        // O valor do service_fee só é incluido para modalidades marketplace/pickup
        if(!empty($order["totals"]["charges"]) && !empty($order["totals"]["charges"]["service_fee"])) {
            if($order["delivery_method"] != "delivery") {
                $deliveryFee += $order["totals"]["charges"]["service_fee"];
            }
        }
            
        if($order["delivery_method"] == "delivery"){
            $order["totals"]["total_order"] += $deliveryFee;
        }
        
        // ------------- Process Products --------------- //
        $isSku = $this->factory->getIdUtilSku();
        $products     = [];
        $ITCOMANDAVEN_SEQ = 0;
        
        unset($item);
        unset($subitems);

        foreach ($order["items"] as $item) {
            $processedProduct = $this->productService->processProduct($generalParams, $item, $nrvendarest, $nrcomanda, $ITCOMANDAVEN_SEQ, $isSku);
            

            if (is_array($processedProduct) && !isset($processedProduct["CDPRODUTO"]))
                $products = array_merge($products, $processedProduct);
            else
                array_push($products, $processedProduct);
        }

        if(empty($products)) {
            throw new \Exception("Ocorreu uma falha na validação de produtos.", 187);
        }

        // ------------- Insert Order Detailment -------- //
        $this->factory->insertSale($nrvendarest, $generalParams, $consumer);
        
        // Removido a pedido da Rappi
        //$OBSGERAL .= !empty($consumer['consumerInfo']['phone_number']) ? (' TEL: '. $consumer['consumerInfo']['phone_number']) : '';

        if($sharedDiscount > 0){
            $this->shareDiscount($sharedDiscount, $products);
        }
        
        if (($order["IDIMPAUTCUPEXP"] == "S") && ($order["IDIMPAUTCUPFOS"] == "S"))  {
            $order["IDIMPCUPEXP"] = 2;
        } else { 
            $order["IDIMPCUPEXP"] = 0;
        }
        
        $this->factory->saveComandaven($generalParams, $order, $deliveryFee, $consumer);
                                      
        $paymentCode = ($order['delivery_method'] == "marketplace") ? $order['payment_method'] : "RAP";
        
        // ------------- Insert Payments ---------------- //

        $this->processPayment($generalParams["CDFILIAL"], $nrvendarest, $order['totals']['total_order'], $generalParams["CDCLIENTE"], $generalParams["CDCONSUMIDOR"], $paymentCode);
        
        
        $this->normalizeDiscountFromThreeDecimals($products);
        
        // ------------- Insert Products ---------------- //
        foreach ($products as $product) {
            if(empty($product['VRPRECCOMVEN']) || (bccomp($product['VRPRECCOMVEN'], "0", 2) <= 0)) {
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui preço inferior a 0.01 centavo. Valor: {$product['VRPRECCOMVEN']}", 191);
            }
            
            if(bccomp($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2) >= 0) {
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui DESCONTO superior ao VALOR: Desconto -> {$product['VRDESCCOMVEN']} Valor -> {$product['VRPRECCOMVEN']}", 192);
            }
            
            $this->productService->insertProduct(array_merge($product, $generalParams));
        }
        
        return [
            'CDFILIAL'      => $generalParams["CDFILIAL"],
            'NRCOMANDA'     => $nrcomanda,
            'NRCOMANDAEXT'  => $order["order_id"],
            'NRVENDAREST'   => $nrvendarest
        ];
    }
    
    private function processPayment($cdfilial, $nrvendarest, $paymentValue, $cdcliente, $cdconsumidor, $receipt)
    {
        
        $nrseqmovdlv = $this->util->newCode("MOVCAIXADLV$cdfilial$nrvendarest", 10);
        
        $paymentCode = $this->paymentFactory->getReceiptType($cdfilial, $receipt);
    
        $this->paymentFactory->saveMovcaixadlv(
            $cdfilial, 
            $nrvendarest, 
            $nrseqmovdlv, 
            $paymentCode, 
            $paymentValue,
            $cdcliente, 
            $cdconsumidor 
        );
    }
    
    public function validateOrder($cdFilial, $nrVendaRest)
    {
        $this->factory->validateOrder($cdFilial, $nrVendaRest);
    }
    
    public function disableIntegration()
    {
        $this->factory->disableIntegration();
    }

    public function getConfiguration()
    {
        return $this->factory->getConfiguration();
    }
    
    /* 
    * Funcao Rateia Descounto sob produtos para adequar os descontos aos produtos
    *
    */
    
    public function shareDiscount($initialDiscount, &$products)
    {
        $idxToApplyDiscount   = [];
        $totalDiscount        = $initialDiscount;
        $discountLeft         = $initialDiscount;
        $productsTotalValue   = "0.00";
        $discountGivenShare   = "0.00";
        $maxValueItem         = "0.00";
        $maxValueIdx          = -1;
        
        foreach($products as &$product) {

            if(!isset($product['VRDESCCOMVEN']) || $product['VRDESCCOMVEN'] == null){
                $product['VRDESCCOMVEN'] = 0;
            } 
            $product['discountShare'] = "0.00";
        }
        
        if(isset($product))
            unset($product);

        foreach($products as $i => $product) {
            if((bccomp($product["VRPRECCOMVEN"], "0.01", 2) == 1)) { 
                $productsTotalValue = bcadd($productsTotalValue, $product["VRPRECCOMVEN"], 2);
                array_push($idxToApplyDiscount, $i);
            }
        }

        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $discountShare = bcdiv((bcmul($products[$idxWithDiscount]["VRPRECCOMVEN"], "100.00", 2)), $productsTotalValue, 2);
            $products[$idxWithDiscount]["discountShare"] = $discountShare;
            $discountGivenShare = bcadd($discountGivenShare, $discountShare, 2);
        }

        foreach($products as $i => $product) {
            if(bccomp($product['discountShare'], $maxValueItem, 2) == 1) {
                $maxValueIdx  = $i;
                $maxValueItem = $product['discountShare'];
            }            
        }

        if($maxValueIdx != -1) {
            $products[$maxValueIdx]["discountShare"] = bcadd($products[$maxValueIdx]["discountShare"], bcsub("100.00", $discountGivenShare, 2), 2);
        }
    
        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $itemDiscount = bcdiv((bcmul($products[$idxWithDiscount]["discountShare"], $totalDiscount, 2)), "100.00", 2);
            $products[$idxWithDiscount]["VRDESCCOMVEN"] += $itemDiscount;

            $discountLeft = bcsub($discountLeft, $itemDiscount, 2);
        }

        if($maxValueIdx != -1) {
            $products[$maxValueIdx]["VRDESCCOMVEN"] = bcadd($products[$maxValueIdx]["VRDESCCOMVEN"], $discountLeft, 2);
            $discountLeft = "0.00";
        }
        
        $leftDiscount = "0.00";
        foreach($products as &$product) {
            if( (bccomp($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2) == 1) || (bccomp($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2) == 0) )  {
                $leftDiscount = bcadd($leftDiscount, bcsub($product['VRDESCCOMVEN'], $product['VRPRECCOMVEN'], 2), 2);
                $leftDiscount = bcadd($leftDiscount, "0.01", 2);
                
                $product['VRDESCCOMVEN'] = bcsub($product['VRDESCCOMVEN'], $leftDiscount, 2);
            }
        }
        
        if(isset($product))
            unset($product);

        foreach($products as &$product){
            $maxAddDisc = (bcsub(bcsub($product['VRPRECCOMVEN'], "0.01", 2), $product['VRDESCCOMVEN'], 2));
            if($maxAddDisc >= $leftDiscount) {
                $addDisc = $leftDiscount;
            } else{
                $addDisc = $maxAddDisc;
            }
            
            $product['VRDESCCOMVEN'] = bcadd($product['VRDESCCOMVEN'], $addDisc, 2);
            $leftDiscount = bcsub($leftDiscount, $addDisc, 2);
        }

        if(isset($product))
            unset($product);
        
        return $products;
    }
    
    /*
     * Calcula o balancete da venda informada pela Rappi
     * 
    */
    private function balanceProducts($products, $deliveryFee)
    {
        return array_reduce($products, function ($carry, $item) {
            return bcadd($carry, $item["VRPRECCOMVEN"], 2);
        }, $deliveryFee);
    }
    
    public function updateAccessToken($newToken)
    {
        return $this->factory->updateAccessToken($newToken);
    }
    
    public function getDeliveryMethodParams($cdFilial)
    {
        return $this->branchFactory->getDeliveryMethodParams($cdFilial); 
    }
    
    public function getDeliveryMethod($deliveryMethod)
    { 
        if ($deliveryMethod == "marketplace") {
            return 'M';
            
        } else if ($deliveryMethod == "delivery") {
            return 'F';
            
        }   
        
        return 'N';
    }
    
    private function bc_floor($value)
    {
        if (strpos($value, '.') !== false) {
            list($integer, $fraction) = explode('.', $value, 2);
            if ((int)$integer >= 0) {
                return bcadd($integer, '0', 0);
            } else {
                return bcsub($integer, '1', 0);
            }
        }
        return $value;
    }
    
    private function normalizeDiscountFromThreeDecimals(&$products) {
        $arrayTemp = [];
        
        foreach ($products as &$product) {
            $tmpValor = (string)$product['VRDESCCOMVEN'];
            $product['VRDESCCOMVEN'] = $tmpValor;
        }
        
        if(isset($product))
            unset($product);
        
        foreach ($products as $key => &$product) {

            if (strpos($product['VRDESCCOMVEN'], '.') !== false && strlen(substr(strrchr($product['VRDESCCOMVEN'], '.'), 1)) > 2) {
                $valorMult100 = bcmul($product['VRDESCCOMVEN'], 100, 2);
                $valorInteiro = bcdiv($valorMult100, 1, 0);
                $sobra        = bcsub($valorMult100, $valorInteiro, 1);

                if (isset($arrayTemp[$product['CDPRODUTO']])) {
                    $arrayTemp[$product['CDPRODUTO']]['value'] = bcadd($arrayTemp[$product['CDPRODUTO']]['value'], $sobra, 2);
                    array_push($arrayTemp[$product['CDPRODUTO']]['ref'], $key);
                }
                
                else {
                    $remaining = [
                        'value' => $sobra,
                        'ref'   => [$key]
                    ];
                    
                    $arrayTemp[$product['CDPRODUTO']] = $remaining;
                }
            }
            
            $product['VRDESCCOMVEN'] = bcmul($product['VRDESCCOMVEN'], '1', 2);
            
        }
        
        if(isset($product))
            unset($product);
        
        foreach($arrayTemp as $tmp) {
            $remaining   = bcdiv($tmp['value'], '100', 2);
            $applied = 0;
            
            foreach($tmp['ref'] as $indexCart) {
                $totalDiscount = bcadd($remaining, bcmul($products[$indexCart]['VRDESCCOMVEN'], '1', 2), 2);

                if(bcsub($products[$indexCart]['VRPRECCOMVEN'], $totalDiscount, 2) >= '0.01') {
                    $products[$indexCart]['VRDESCCOMVEN'] = $totalDiscount;
                    $applied = $remaining;
                    break;
                }
            }
            
            if(bccomp($remaining, $applied) != 0) {
                 throw new \Exception("Falha no rateio de desconto cumulativo");
            }
        }
    }
    
}
